import os
from typing import *

import requests

from parsee.templates.job_template import JobTemplate
from parsee.templates.template_from_json import from_json_dict
from parsee.extraction.extractor_elements import StandardDocumentFormat
from parsee.converters.json_to_raw import load_document_from_json
from parsee.extraction.extractor_dataclasses import ParseeAnswer, ParseeMeta, source_from_json, AssignedAnswer
from parsee.extraction.extractor_dataclasses import Base64Image
from parsee.converters.image_creation import from_bytes


class ParseeCloud:

    def __init__(self, api_key: Optional[str], custom_host: Optional[str] = None):

        self.api_key = api_key
        self.host = custom_host if custom_host is not None else (os.getenv("BACKEND_HOST") if os.getenv("BACKEND_HOST") is not None else "https://backend.parsee.ai")

    def _headers(self):
        return {"Authorization": self.api_key}

    def get_template(self, template_id: str) -> JobTemplate:
        url = f"{self.host}/{'public/' if self.api_key is None else ''}api/extraction/template/id/{template_id}"
        template_request = requests.get(url, headers=self._headers())
        if template_request.content == b'':
            raise Exception("template not found or not accessible")
        return from_json_dict(template_request.json())

    def get_document(self, source_identifier: str) -> StandardDocumentFormat:
        url = f"{self.host}/api/document/get-json-document?identifier={source_identifier}"
        data = requests.get(url, headers=self._headers()).json()
        return load_document_from_json(data)

    def save_template(self, template: JobTemplate, public: bool = False) -> str:
        url = f"{self.host}/api/extraction/template"
        template_json_dict = template.to_json_dict()
        template_json_dict = {**template_json_dict, "public": public}
        request = requests.post(url, json=template_json_dict, headers=self._headers())
        return request.text

    # currently only output for general questions is supported here (i.e. textual output)
    def get_output(self, source_identifier: str, template_id: str) -> List[ParseeAnswer]:
        url = f"{self.host}/api/extraction/output/json/{source_identifier}"
        data = requests.get(url, headers=self._headers()).json()
        output = []
        for entry in data:
            if entry["templateId"] == template_id and entry["text"] is not None:
                class_id = entry["detectedClass"]
                class_value = entry["text"]["answerParsed"]
                meta_answers = []
                for item in entry["text"]["metaInfo"]:
                    meta_answers.append(ParseeMeta(item["model"], 0, [], item["class_id"], item["value"], item["prob"]))
                sources = [source_from_json(x) for x in entry["sources"]]
                output.append(ParseeAnswer(entry["model"], sources, class_id, class_value, "", True, meta_answers))
        return output

    def upload_file(self, file_path: str) -> str:
        """
        upload a PDF, HTML or image file to Parsee Cloud
        """
        with open(file_path, "rb") as f:
            data = f.read()
        file_data = {'file': (os.path.basename(file_path), data)}

        extractor_url = f"{self.host}/api/document/upload?method=simple"

        r = requests.post(extractor_url, {}, files=file_data, headers=self._headers())

        data = r.json()

        return list(data[0].values())[0]

    def add_assigned_answers(self, template_id: str, source_identifier: str, answers: List[AssignedAnswer]) -> bool:
        """
        adds one or more assigned answers to the Parsee Cloud database
        """
        failed = False
        for answer in answers:
            data = {
                "sourceIdentifier": source_identifier,
                "templateId": template_id,
                "itemId": answer.class_id,
                "newValue": answer.class_value,
                "newMeta": [
                    {"model": "manual", "class_id": meta_item.class_id, "value": meta_item.class_value, "prob": 1.0} for meta_item in answer.meta
                ],
                "sources": [source.to_json_dict() for source in answer.sources]
            }

            url = f"{self.host}/api/extraction/output/general-query"

            r = requests.post(url, json=data, headers=self._headers())

            if r.status_code != 200:
                failed = True
        return not failed

    def _get_image(self, source_identifier: str, page_index: int) -> bytes:

        url = f"{self.host}/api/document/images?identifier={source_identifier}&page-index={page_index}"

        return requests.get(url, headers=self._headers()).content

    def get_image(self, source_identifier: str, page_index: int, max_image_size: int = 2000) -> Base64Image:

        data = self._get_image(source_identifier, page_index)

        return from_bytes(data, "image/jpeg", max_image_size)

    def get_image_and_save(self, source_identifier: str, page_index: int, output_path: str):
        """
        :param output_path: Output image is always a JPEG.
        """

        with open(output_path, "wb") as f:
            f.write(self._get_image(source_identifier, page_index))


import json
from typing import List, Optional, Dict, Union, Any
from pydantic.dataclasses import dataclass

from parsee.utils.enums import OutputType, ContextType, SearchStrategy


@dataclass
class StructuringItemSchema:
    type: OutputType
    context: ContextType
    id: str
    title: str
    additionalInfo: str
    model: Optional[str]
    searchStrategy: SearchStrategy
    example: Union[None, str]
    keywords: Union[None, str]
    valuesList: Union[None, List[str]]
    defaultValue: Union[None, str]
    customArgsJson: Union[None, str]

    def parsed_json_args(self) -> any:
        if self.customArgsJson is None:
            return None
        try:
            return json.loads(self.customArgsJson)
        except Exception as e:
            return None

    def __hash__(self):
        return hash(self.id)

    def to_json_dict(self) -> Dict:
        return {"type": self.type.value, "context": self.context.value, "id": self.id, "title": self.title, "additionalInfo": self.additionalInfo,
                "model": self.model, "searchStrategy": self.searchStrategy.value, "example": self.example, "keywords": self.keywords, "valuesList": self.valuesList,
                "defaultValue": self.defaultValue, "customArgsJson": self.customArgsJson
                }


@dataclass
class StructuringSchema:
    items: List[StructuringItemSchema]
    settings: Dict[str, Any]

    def to_json_dict(self) -> Dict:
        return {"items": [x.to_json_dict for x in self.items], "settings": self.settings}


@dataclass
class GeneralQueryItemSchema (StructuringItemSchema):
    metaInfoIds: List[str]

    def to_json_dict(self) -> Dict:
        return {**super().to_json_dict(), "metaInfoIds": self.metaInfoIds}


@dataclass
class GeneralQuerySchema:
    items: List[GeneralQueryItemSchema]
    settings: Dict[str, Any]

    def to_json_dict(self) -> Dict:
        return {"items": [x.to_json_dict() for x in self.items], "settings": self.settings}



from typing import *

from parsee.datasets.dataset_dataclasses import DatasetRow
from parsee.extraction.extractor_dataclasses import Base64Image


class Prompt:

    def __init__(self, description: str, main_task: str, additional_info: str, full_example: str, available_data: Union[str, List[Base64Image]]):

        self.description = description
        self.main_task = main_task
        self.additional_info = additional_info
        self.full_example = full_example
        self.available_data = available_data

    def __str__(self) -> str:
        return f'''{self.description} \n

                    {self.main_task} \n
                    
                    {self.additional_info} \n

                    {self.full_example} \n
                    
                    {self.available_data if isinstance(self.available_data, str) else ''}
                    '''



from __future__ import annotations

import numpy as np
from dataclasses import dataclass
import datetime
from typing import List, Dict, Tuple, Union, Optional
from hashlib import sha256
from decimal import Decimal

from parsee.utils.enums import DocumentType


@dataclass
class ParseeBucket:
    schema_id: str
    bucket_id: str
    li_identifier: str
    model: str
    prob: float
    kv_index: int
    definition_idx: int
    multiplier: Decimal

    def dict_json(self) -> Dict:
        return {"schema_id": self.schema_id, "bucket_id": self.bucket_id, "li_identifier": self.li_identifier, "prob": self.prob, "kv_index": self.kv_index, "definition_idx": self.definition_idx, "multiplier": float(self.multiplier)}


@dataclass
class AssignedBucket:
    bucket_id: str
    kv_index: int
    definition_index: int
    multiplier: Decimal


@dataclass
class CoreExtractionRequest:
    source_type: str
    source_identifier: str


@dataclass
class ExtractedSource:
    source_type: DocumentType
    coordinates: Union[None, Dict]
    xpath: Union[None, str]
    element_index: int
    other_info: Union[Dict, None]

    def to_json_dict(self):
        return {"source_type": self.source_type.value, "coordinates": self.coordinates, "xpath": self.xpath, "element_index": self.element_index, "other_info": self.other_info}

    def identifier(self):
        return self.element_index

    def to_location_id(self) -> str:
        key = "coordinates" if self.coordinates is not None else "xpath"
        return sha256(str(self.to_json_dict()[key]).encode('utf-8')).hexdigest()


def source_from_json(json_dict: Dict) -> ExtractedSource:
    return ExtractedSource(DocumentType(json_dict["source_type"]), json_dict["coordinates"], json_dict["xpath"], json_dict["element_index"], json_dict["other_info"])


@dataclass
class ParseeLocation:
    model: str
    partial_prob: float
    detected_class: str
    prob: float
    source: ExtractedSource
    meta: List[ParseeMeta]


@dataclass
class AssignedLocation:
    class_id: str
    element_index: int
    is_partial: bool
    meta: List[AssignedMeta]
    mappings: List[AssignedBucket]


@dataclass
class ParseeMeta:
    model: str
    column_index: int
    source: List[ExtractedSource]
    class_id: str
    class_value: str
    prob: float

    def to_json_dict(self):
        return {"model": self.model, "class_id": self.class_id, "value": self.class_value, "prob": self.prob}


def meta_from_json(json_dict: Dict, col_idx: int, source: List[ExtractedSource]) -> ParseeMeta:
    return ParseeMeta(json_dict["model"], col_idx, source, json_dict["class_id"], json_dict["value"], json_dict["prob"])


@dataclass
class AssignedMeta:
    class_id: str
    class_value: str
    column_index: Optional[int] = None


@dataclass
class ParseeAnswer:
    model: str
    sources: List[ExtractedSource]
    class_id: str
    class_value: str
    raw_value: str
    parse_successful: bool
    meta: List[ParseeMeta]

    def to_json_dict(self):
        return {"class_id": self.class_id, "parsed": self.class_value, "raw": self.raw_value, "parseSuccessful": self.parse_successful, "meta": [x.to_json_dict() for x in self.meta], "sources": [x.to_json_dict() for x in self.sources]}

    def meta_hash(self) -> str:
        relevant_values = {"class_id": self.class_id, "meta": [x.to_json_dict() for x in self.meta]}
        return sha256(str(relevant_values).encode('utf-8')).hexdigest()

    def meta_key(self) -> str:
        if len(self.meta) == 0:
            return ""
        meta_sorted = sorted(self.meta, key=lambda x: x.class_id)
        return "_".join([f"{x.class_id}:{x.class_value}" for x in meta_sorted])

    def unique_id(self) -> str:
        return f"{self.class_id}:{self.meta_key()}"

    def to_assigned_answer(self) -> AssignedAnswer:
        return AssignedAnswer(self.class_id, self.class_value, [AssignedMeta(x.class_id, x.class_value) for x in self.meta], self.sources)


@dataclass
class AssignedAnswer:
    class_id: str
    class_value: str
    meta: List[AssignedMeta]
    sources: List[ExtractedSource]


@dataclass
class Base64Image:
    media_type: str
    data: str



import json
from typing import List

from parsee.extraction.extractor_elements import ExtractedEl, StructuredTable, StructuredTableCell, StructuredRow, StandardDocumentFormat
from parsee.utils.enums import ElementType, DocumentType
from parsee.extraction.extractor_dataclasses import ExtractedSource


def source_from_json(json_dict) -> ExtractedSource:
    return ExtractedSource(DocumentType(json_dict["source_type"]), json_dict["coordinates"], json_dict["xpath"], json_dict["element_index"], json_dict["other_info"])


def load_elements_from_json(elements_dict_list: List) -> List[ExtractedEl]:

    output: List[ExtractedEl] = []
    for el in elements_dict_list:

        if el["el_type"] == "text":
            output.append(ExtractedEl(ElementType.TEXT, source_from_json(el["source"]), el["text"]))
        elif el["el_type"] == "figure":
            output.append(ExtractedEl(ElementType.FIGURE, source_from_json(el["source"]), None))
        else:
            source = source_from_json(el["source"])
            rows_structured = []
            for row in el["rows"]:
                row_type = row['row_type']
                values_structured = []
                for val in row['values']:
                    values_structured.append(StructuredTableCell(val["val"], val["colspan"]))
                rows_structured.append(StructuredRow(row_type, values_structured))
            output.append(StructuredTable(source, rows_structured))
    return output


def load_document_from_json(json_contents) -> StandardDocumentFormat:

    source_type = DocumentType(json_contents["source_type"])
    source_identifier = json_contents["source_identifier"]
    elements = load_elements_from_json(json_contents["elements"])

    return StandardDocumentFormat(source_type, source_identifier, elements, None)



import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

import openai
import tiktoken

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, get_tokens_encoded, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.extractor_dataclasses import Base64Image


class ChatGPTModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer
        openai.api_key = model.api_key if model.api_key is not None else os.getenv("OPENAI_KEY")

    def _call_api(self, prompt: str, images: List[Base64Image], retries: int = 0, wait: int = 5) -> Tuple[str, Decimal]:
        user_message_content = [
            {
                "type": "text",
                "text": prompt
            }
        ]
        if len(images) > 0:
            user_message_content += [
                {
                    "type": "text",
                    "text": "These are the images:"
                }
            ]
            user_message_content += [
                {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:{x.media_type};base64,{x.data}"
                    }
                }
                for x in images
            ]
        try:
            response = openai.ChatCompletion.create(
                model=self.spec.internal_name,
                messages=[
                    {"role": "user", "content": user_message_content}
                ],
                temperature=0,
                max_tokens=self.max_tokens_answer,
                top_p=1,
                frequency_penalty=0,
                presence_penalty=0
            )

            answer = response['choices'][0]["message"]["content"]
            final_cost = Decimal(int(response["usage"]['total_tokens']) * (self.spec.price_per_1k_tokens/1000)) if self.spec.price_per_1k_tokens is not None else Decimal(0)
            return answer, final_cost
        except Exception as e:
            if retries < self.max_retries:
                time.sleep(wait * 2 ** retries)
                return self._call_api(prompt, images, retries + 1)
            else:
                return "", Decimal(0)

    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, _ = truncate_prompt(str(prompt), self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt, prompt.available_data if self.spec.multimodal else [])



from decimal import Decimal
from functools import reduce
from typing import *

import tiktoken

from parsee.extraction.extractor_elements import StandardDocumentFormat
from parsee.templates.job_template import JobTemplate
from parsee.storage.interfaces import StorageManager
from parsee.storage.in_memory_storage import InMemoryStorageManager
from parsee.extraction.models.model_loader import question_models_from_schema, ModelLoader
from parsee.extraction.tasks.questions.features import GeneralQueriesPromptBuilder
from parsee.extraction.models.llm_models.llm_base_model import truncate_prompt
from parsee.datasets.dataset_dataclasses import DatasetRow
from parsee.extraction.extractor_dataclasses import AssignedAnswer


def create_dataset_rows(template: JobTemplate, document: StandardDocumentFormat, assigned_answers: List[AssignedAnswer], storage: Optional[StorageManager] = None, max_tokens_prompt=4000, custom_model_loader: Optional[ModelLoader] = None) -> List[DatasetRow]:
    encoding = tiktoken.get_encoding("cl100k_base")
    storage = InMemoryStorageManager(None) if storage is None else storage
    model_loader = ModelLoader(storage) if custom_model_loader is None else custom_model_loader
    template = storage.db_values_template(template, False)
    question_feature_builder = GeneralQueriesPromptBuilder(storage)
    question_models = question_models_from_schema(template.questions, template.meta, model_loader, {"truth_questions": assigned_answers})
    question_rows = []
    if len(question_models) > 0:
        question_model = question_models[0]
        answers = question_model.predict_answers(document)
        if len(answers) > 0:
            for item in question_model.items:
                meta_items_filtered = [x for x in template.meta if x.id in item.metaInfoIds]
                relevant_elements = question_feature_builder.get_relevant_elements(item, document)
                prompt = question_feature_builder.build_prompt(item, meta_items_filtered, document, relevant_elements)
                real_prompt, _ = truncate_prompt(str(prompt), encoding, max_tokens_prompt)
                answers_filtered = [x for x in answers if x.class_id == item.id]
                if len(answers_filtered) > 0:
                    # join together in case the answer has multiple blocks
                    full_answer = question_feature_builder.build_raw_value(answers_filtered, item, meta_items_filtered)
                    row = DatasetRow(document.source_identifier, template.id, item.id, {"full_prompt": real_prompt})
                    row.assign_truth_values({"assigned": full_answer})
                    question_rows.append(row)
    return question_rows



from typing import *

from parsee.extraction.extractor_elements import StandardDocumentFormat
from parsee.templates.general_structuring_schema import StructuringItemSchema, GeneralQueryItemSchema
from parsee.extraction.extractor_dataclasses import ParseeAnswer, AssignedAnswer, ParseeMeta


class QuestionModel:

    model_name = ""

    def __init__(self, items: List[GeneralQueryItemSchema], meta_items: List[StructuringItemSchema]):
        self.items = items
        self.meta = meta_items

    def predict_answers(self, document: StandardDocumentFormat) -> List[ParseeAnswer]:
        raise NotImplemented


class AssignedQuestionModel(QuestionModel):

    model_name = "manual"

    def __init__(self, items: List[GeneralQueryItemSchema], meta_items: List[StructuringItemSchema], truth_questions: List[AssignedAnswer], **kwargs):
        super().__init__(items, meta_items)
        self.answers = truth_questions

    def predict_answers(self, document: StandardDocumentFormat) -> List[ParseeAnswer]:
        output: List[ParseeAnswer] = []

        for answer in self.answers:
            meta_values = [ParseeMeta(self.model_name, meta.column_index if meta.column_index is not None else 0, answer.sources, meta.class_id, meta.class_value, 1) for meta in answer.meta]
            output.append(ParseeAnswer(self.model_name, answer.sources, answer.class_id, answer.class_value, "", True, meta_values))

        return output